import requests
import base64
import datetime
import traceback
import json

############################# NEXUS REST API V2 Functions ############################
class NEXUSIC_REST():
    """
    NEXUS IC REST API class allows the user to communicate with Wood NEXUS IC REST API using python.

    Prerequisites:
    -------------
        - Python > 3.7
        - NEXUS IC > V6.6
        - IC-Web > V6.6

    NEXUS IC Documentation:
    ----------------------
        The NEXUS IC REST API documentation can be found in the below link:
        https://docs.nexusic.com/6.6/ic-web.rest.v2.html

        A specific NEXUS IC version can be specified in the above link by changing **6.6** to the desired NEXUS IC version*
    """

    _error_msgs = ['An existing connection was forcibly closed by the remote host',
                   'Remote end closed connection without response']

    def __init__(self, icweb_uri, authentication_type='APIKEY',
                 username=None, password=None, api_key=None,
                 max_attempts=1, timeout=None, verbose=False, verify=True):
        """

            :param icweb_uri: (``string``) - IC-Web URL.
            :param authentication_type: (``string`` - optional) - This can be one of the following values  (default value APIKEY).

                .. code-block:: python

                        ['APIKEY', 'BASIC']

            :param username: (``string`` - optional) - Default value None.
            :param password: (``string`` - optional) - Default value None.
            :param api_key: (``string`` - optional) - Default value None.
            :param max_attempts: (``int`` - optional) - Maximum number of attempts if disconnected (default value 1).
            :param timeout: (``int`` - optional) - Timeout threshold in seconds (default value None).
            :param verbose: (``bool`` - optional) - Print internal messages if True (default value False).
            :param verify: (``bool`` - optional) - By pass SSL verification if True (default value True).

            :returns: None
        """

        self.icweb_uri = icweb_uri
        self.authentication_type = authentication_type

        self.api_key = api_key
        self.username = username
        self.password = password

        self.max_attempts = max_attempts
        self.timeout = timeout   # TODO: To be added to the REST calls
        self.verbose = verbose
        self.verify = verify

        assert authentication_type in ['APIKEY', 'BASIC'], 'Incorrect authentication type'

        if authentication_type == 'APIKEY':
            if self.api_key != None:
                self.key_64 = self.generate_base64(api_key)
            else:
                raise Exception('API Key is not valid, please provide a valid API Key')
        elif authentication_type == 'BASIC':
            if self.username != None or self.password != None:
                self.key_64 = self.generate_base64(self.username + ':' + self.password)
            else:
                raise Exception('Username and/or password are not valid, please provide a valid username/password')
        else:
            raise Exception('Authentication type was not specified')

        self.hash = self.generate_hash()

        # Get current NEXUS version
        version, version_status_code = self.getVersion()
        self.version = version['version'].split('.')
        self.schema = version['schema'].split('.')

    ################################ Core REST API Calls #################################
    ######################################################################################
    def generate_base64(self, value):
        """
            Generate base64 string

            :param value: (``string``) - String value to be converted to base64.

            :returns: (``String``) - base64 string.
        """
        return str(base64.b64encode(bytes(value, 'utf-8')), "utf-8")

    ######################################################################################
    def generate_hash(self, verbose=False):
        """
            Generate hash key to be used in the REST calls

            :param verbose: (``bool`` - optional) - Print internal messages if True (default value False).

            :return: (``string``) - Hash key
        """

        result, result_code = self.authenticate(verbose=verbose)

        if result_code == 200:
            return result.get('hash')
        else:
            errorMsg = traceback.format_exc()
            raise Exception(result + '\n' + errorMsg)

    ######################################################################################
    def addParamToURI(self, uri, param, value):
        if '?' in uri:
            newUri = uri + '&' + param + '=' + value
        else:
            newUri = uri + '?' + param + '=' + value

        return newUri

    ######################################################################################
    def validate_and_return_response(self, response, message, raw=False):
        """
            Validate response by comparing it against the acceptable status codes. Returns response and status code.

            :param response: (``requests.response``) - The response from the query.
            :param message: (``string`` - optional) - Desired custom error message.
            :param raw: (``bool`` - optional) - Defines whether to return response raw format or json (default value False).

            :return: (``tuple`` - dict or string, string) - response (json, raw or error string) and response status code.
        """

        if response.status_code in [200]:
            if raw:
                return response.raw, response.status_code
            else:
                return response.json(), response.status_code
        else:
            return str(message) + str(response.status_code) + ': ' + str(response.text), \
                   response.status_code

    ######################################################################################
    def authenticate(self, verbose=False):
        """
            Authenticate with NEXUS IC using the defined authentication type used in the class constructor.
            For more details see https://docs.nexusic.com/6.6/ic-web.rest.security.login.html#ic-web-rest-security-login

            :param verbose: (``bool`` - optional) - Print internal messages if True (default value False).

            :return: (``tuple`` - dict or string, string) - response (json, raw or error string) and response status code.
        """

        if self.verbose or verbose:
            print('Authenticating with NEXUS IC...')

        # Adding /data/icweb.dll part to the baseURI
        baseURI = self.icweb_uri + '/data/icweb.dll'

        uri = baseURI + '/security/login'
        res = requests.get(uri, headers={'Authorization': self.authentication_type + ' ' + self.key_64},
                           verify=self.verify)

        return self.validate_and_return_response(res, 'Authentication error ')

    ######################################################################################
    def getVersion(self, current_attempt=1, verbose=False):
        if self.verbose or verbose:
            print('Getting NEXUS IC version...')

        # Adding /data/icweb.dll part to the baseURI
        baseURI = self.icweb_uri + '/data/icweb.dll'

        uri = baseURI + '/version' + '?hash=' + self.hash

        try:
            res = requests.get(uri, verify=self.verify)
            result, result_code = self.validate_and_return_response(res, 'Get version error ')
            return result, result_code
        except Exception as e:
            for error_msg in NEXUSIC_REST._error_msgs:
                if error_msg in str(e):
                    current_attempt += 1
                    if (current_attempt <= self.max_attempts) and not (self.key_64 == None):
                        self.hash = self.generate_hash()
                        return self.getVersion(current_attempt=current_attempt)

            errorMsg = traceback.format_exc()
            raise Exception('Number of attempts: ' + str(current_attempt) + '\n' + errorMsg)

    ######################################################################################
    def getTable(self, tableDBName, xFilter=None, pageSize=None,
                 current_attempt=1, verbose=False):
        """
            Execute GET REST call to get data from specific table

            :param tableDBName: (``string``) - Table name as specified in the database (not the NEXUS IC display table name).
            :param xFilter: ( ``dict`` in ``JSON`` fomrat - optional) - Used to filter data from the required table (default value None).
            :param pageSize: (``int`` - optional) - Page size for response (default value None). When default is used the default NEXUS IC page size will be used (100).
            :param current_attempt: (``int`` - **don't use**) - This arugment is intended for internal method use only.
            :param verbose: (``bool`` - optional) - Print internal messages if True (default value False).

            :return: (``tuple`` - dict or string, string) - response (json, raw or error string) and response status code.
        """

        if self.verbose or verbose:
            print('Getting ' + str(tableDBName) + '...')

        # Adding /data/icweb.dll part to the baseURI
        baseURI = self.icweb_uri + '/data/icweb.dll'

        uri = baseURI + '/bo/' + tableDBName + '/'

        # Adding pageSize
        if pageSize == None:
            pageSize = 100

        uri = self.addParamToURI(uri, 'pageSize', str(pageSize))

        # Adding startRow
        startRow = 0
        uri = self.addParamToURI(uri, 'startRow', str(startRow))

        # Adding hash value
        uri = self.addParamToURI(uri, 'hash', self.hash)

        try:
            if xFilter != None:
                res = requests.get(uri, headers={'X-NEXUS-Filter': xFilter}, verify=self.verify)
            else:
                res = requests.get(uri, verify=self.verify)

            result, result_code = self.validate_and_return_response(res, 'Get ' + tableDBName + ' table error ')

            # Get next page
            uri = uri.replace('startRow=' + str(startRow), 'startRow=' + str(pageSize))
            startRow = pageSize
            while len(result['rows']) < result['totalRows']:
                if xFilter != None:
                    res = requests.get(uri, headers={'X-NEXUS-Filter': xFilter}, verify=self.verify)
                else:
                    res = requests.get(uri, verify=self.verify)

                result2, result_code2 = self.validate_and_return_response(res, 'Get ' + tableDBName + ' table error ')
                result['rows'].extend(result2['rows'])

                prevStartRow = startRow
                startRow += pageSize
                uri = uri.replace('startRow=' + str(prevStartRow), 'startRow=' + str(startRow))

            return result, result_code
        except Exception as e:
            for error_msg in NEXUSIC_REST._error_msgs:
                if error_msg in str(e):
                    current_attempt += 1
                    if (current_attempt <= self.max_attempts) and not (self.key_64 == None):
                        self.hash = self.generate_hash()
                        return self.getTable(tableDBName, xFilter=xFilter, pageSize=pageSize,
                                        current_attempt=current_attempt)

            errorMsg = traceback.format_exc()
            raise Exception('Number of attempts: ' + str(current_attempt) + '\n' + errorMsg)

    ######################################################################################
    def deleteRecord(self, tableDBName, keyValue, current_attempt=1, verbose=False):
        if self.verbose or verbose:
            print('Deleting from ' + str(tableDBName) + '...')

        # Adding /data/icweb.dll part to the baseURI
        baseURI = self.icweb_uri + '/data/icweb.dll'

        uri = baseURI + '/bo/' + tableDBName + '/' + str(keyValue) + '/' + '?hash=' + self.hash

        try:
            res = requests.delete(uri, verify=self.verify)
            result, result_code = self.validate_and_return_response(res, 'Delete ' + tableDBName + ' table error ')
            return result, result_code
        except Exception as e:
            for error_msg in NEXUSIC_REST._error_msgs:
                if error_msg in str(e):
                    current_attempt += 1
                    if (current_attempt <= self.max_attempts) and not (self.key_64 == None):
                        self.hash = self.generate_hash()
                        return self.deleteRecord(tableDBName, keyValue, current_attempt=current_attempt)

            errorMsg = traceback.format_exc()
            raise Exception('Number of attempts: ' + str(current_attempt) + '\n' + errorMsg)

    ######################################################################################
    def getMultimedia(self, rd_id, current_attempt=1, verbose=False):
        if self.verbose or verbose:
            print('Getting multimedia: ' + str(rd_id) + '...')

        # Adding /data/icweb.dll part to the baseURI
        baseURI = self.icweb_uri + '/data/icweb.dll'

        xFilter = str(rd_id) + '/File_Data'
        uri = baseURI + '/bo/' + 'Repository_Data' + '/' + xFilter + '/' + '?hash=' + self.hash

        try:
            res = requests.get(uri, stream=True, verify=self.verify)
            result, result_code = self.validate_and_return_response(res, 'Get multimeida error ', raw=True)
            return result, result_code
        except Exception as e:
            for error_msg in NEXUSIC_REST._error_msgs:
                if error_msg in str(e):
                    current_attempt += 1
                    if (current_attempt <= self.max_attempts) and not (self.key_64 == None):
                        self.hash = self.generate_hash()
                        return self.getMultimedia(rd_id, current_attempt=current_attempt)

            errorMsg = traceback.format_exc()
            raise Exception('Number of attempts: ' + str(current_attempt) + '\n' + errorMsg)

    ######################################################################################
    # V6.6 only
    def getDashboard(self, dashboard_Name, current_attempt=1, verbose=False):
        # Check if minor version is 6
        if not(int(self.version[1]) >= 6):
            return 'This function is not supported in the current NEXUS IC version', 404

        if self.verbose or verbose:
            print('Generating NEXUS IC dashboard...')

        # Get RT_ID
        xFilter = '{"where": [{"field": "Name", "value": "' + dashboard_Name + '"}]}'
        report_json, report_status = self.getTable('Report_Template', xFilter=xFilter)

        if report_status == 404:
            return str(report_status) + ': ' + str(report_json), report_status
        else:
            rt_id = report_json['rows'][0]['RT_ID']

        # Adding /data/icweb.dll part to the baseURI
        baseURI = self.icweb_uri + '/data/icweb.dll'

        uri = baseURI + '/dashboard/' + str(rt_id) + '/' + '?hash=' + self.hash

        try:
            res = requests.get(uri, verify=self.verify)
            result, result_code = self.validate_and_return_response(res, 'Get ' + dashboard_Name + ' error ')
            return result, result_code
        except Exception as e:
            for error_msg in NEXUSIC_REST._error_msgs:
                if error_msg in str(e):
                    current_attempt += 1
                    if (current_attempt <= self.max_attempts) and not (self.key_64 == None):
                        self.hash = self.generate_hash()
                        return self.getDashboard(dashboard_Name, current_attempt=current_attempt)

            errorMsg = traceback.format_exc()
            raise Exception('Number of attempts: ' + str(current_attempt) + '\n' + errorMsg)

    ######################################################################################
    def generateReport(self, report_name, recipient, format='XLSX', current_attempt=1,
                       verbose=False):
        if self.verbose or verbose:
            print('Generating NEXUS IC report...')

        # Get RT_ID
        xFilter = '{"where": [{"field": "Name", "value": "' + report_name + '"}]}'
        report_json, report_status = self.getTable('Report_Template', xFilter=xFilter)

        if report_status == 404:
            return str(report_status) + ': ' + str(report_json), report_status
        else:
            rt_id = report_json['rows'][0]['RT_ID']

        # Adding /data/icweb.dll part to the baseURI
        baseURI = self.icweb_uri + '/data/icweb.dll'

        # Generate report
        uri = baseURI + '/web/generateReport'
        uri += '?key=' + str(rt_id) + '&format=' + format + '&recipient=' + recipient + '&hash=' + self.hash

        try:
            res = requests.post(uri, verify=self.verify)
            result, result_code = self.validate_and_return_response(res, 'Generate ' + report_name + ' report error ')
            return result, result_code
        except Exception as e:
            for error_msg in NEXUSIC_REST._error_msgs:
                if error_msg in str(e):
                    current_attempt += 1
                    if (current_attempt <= self.max_attempts) and not (self.key_64 == None):
                        self.hash = self.generate_hash()
                        return self.generateReport(report_name, recipient, format=format,
                                              current_attempt=current_attempt)

            errorMsg = traceback.format_exc()
            raise Exception('Number of attempts: ' + str(current_attempt) + '\n' + errorMsg)

    ######################################################################################
    def execFunction(self, functionName, parameters=None, current_attempt=1,
                     verbose=False):
        if self.verbose or verbose:
            print('Executing ' + str(functionName) + '...')

        # Adding /data/icweb.dll part to the baseURI
        baseURI = self.icweb_uri + '/data/icweb.dll'

        uri = baseURI + '/function/' + functionName
        uri += '/' + '?hash=' + self.hash

        try:
            res = requests.post(uri, json=parameters, verify=self.verify)
            result, result_code = self.validate_and_return_response(res, 'Execute function ' + functionName + ' error ')
            return result, result_code
        except Exception as e:
            for error_msg in NEXUSIC_REST._error_msgs:
                if error_msg in str(e):
                    current_attempt += 1
                    if (current_attempt <= self.max_attempts) and not (self.key_64 == None):
                        self.hash = self.generate_hash()
                        return self.execFunction(functionName, parameters=parameters,
                                            current_attempt=current_attempt)

            errorMsg = traceback.format_exc()
            raise Exception('Number of attempts: ' + str(current_attempt) + '\n' + errorMsg)

    ######################################################################################
    def execUpdate(self, tableDBName, tableID, body, current_attempt=1, verbose=False):
        if self.verbose or verbose:
            print('Updating ' + str(tableDBName) + '...')

        # Adding /data/icweb.dll part to the baseURI
        baseURI = self.icweb_uri + '/data/icweb.dll'

        uri = baseURI + '/bo/' + tableDBName + '/' + tableID
        uri += '?hash=' + self.hash

        try:
            res = requests.post(uri, json=body, verify=self.verify)
            result, result_code = self.validate_and_return_response(res, 'Updating ' + tableDBName + ' error ')
            return result, result_code
        except Exception as e:
            for error_msg in NEXUSIC_REST._error_msgs:
                if error_msg in str(e):
                    current_attempt += 1
                    if (current_attempt <= self.max_attempts) and not (self.key_64 == None):
                        self.hash = self.generate_hash()
                        return self.execUpdate(tableDBName, tableID, body,
                                          current_attempt=current_attempt)

            errorMsg = traceback.format_exc()
            raise Exception('Number of attempts: ' + str(current_attempt) + '\n' + errorMsg)

    ######################################################################################
    def createNewRecord(self, tableDBName, body, key_value=0, current_attempt=1, verbose=False):
        if self.verbose or verbose:
            print('Creating new record in ' + str(tableDBName) + '...')

        # Adding /data/icweb.dll part to the baseURI
        baseURI = self.icweb_uri + '/data/icweb.dll'

        uri = baseURI + '/bo/' + tableDBName + '/' + str(key_value)
        uri += '?hash=' + self.hash

        try:
            res = requests.put(uri, json=body, verify=self.verify)
            result, result_code = self.validate_and_return_response(res, 'Creating new record in ' + tableDBName + ' error ')
            return result, result_code
        except Exception as e:
            for error_msg in NEXUSIC_REST._error_msgs:
                if error_msg in str(e):
                    current_attempt += 1
                    if (current_attempt <= self.max_attempts) and not (self.key_64 == None):
                        self.hash = self.generate_hash()
                        return self.execUpdate(tableDBName, body, key_value=key_value,
                                               current_attempt=current_attempt)

            errorMsg = traceback.format_exc()
            raise Exception('Number of attempts: ' + str(current_attempt) + '\n' + errorMsg)

    ############################ Specific REST API Calls #################################
    ######################################################################################
    def getAssetLocationByName(self, assetName, assetView,
                         pageSize=None, verbose=False):
        x_filter = {"operator": "and",
                    "where": [{"field": "Comp_View.Name", "value": assetView},
                              {"field": "Component.Name", "value": assetName}]
                    }

        x_filter = json.dumps(x_filter)

        result, result_code = self.getTable('View_Node', xFilter=x_filter, pageSize=pageSize, verbose=verbose)

        return result, result_code

    ######################################################################################
    def getAssetLocationByID(self, assetID, assetView,
                         pageSize=None, verbose=False):
        x_filter = {"operator": "and",
                    "where": [{"field": "Comp_View.Name", "value": assetView},
                              {"field": "Component.Component_ID", "value": assetID}]
                    }

        x_filter = json.dumps(x_filter)

        result, result_code = self.getTable('View_Node', xFilter=x_filter, pageSize=pageSize, verbose=verbose)

        return result, result_code

    ######################################################################################
    def getAssetChildren(self, assetLocation, assetView=None, searchType='MAX LEVEL', assetTypes=None,
                         maxLevel=-1, atLevel=-1, pageSize=None, verbose=False):
        assert searchType in ['MAX LEVEL', 'AT LEVEL'], 'Incorrect search type'

        x_filter = {"operator": "and",
                    "where": [{"field": "View_Node.VN_ID", "method": "ch", "value": assetLocation}]
                    }

        # Apply asset view filter
        if assetView != None:
            x_filter['where'].append({"field": "Comp_View.Name", "value": assetView})

        # Apply asset type filters
        if assetTypes != None:
            # Get asset type IDs
            assetTypeIDs = self.getAssetTypesID(assetTypes, pageSize=pageSize, verbose=verbose)

            # Asset children filter
            x_filter['where'].append({"field": "CT_ID", "method": "in", "items": assetTypeIDs})

        # Apply level filter
        if searchType == 'MAX LEVEL' and maxLevel != -1:
            x_filter['where'].append({"field": "Level", "method": "le", "value": maxLevel})
        elif searchType == 'AT LEVEL' and atLevel != -1:
            x_filter['where'].append({"field": "Level", "value": atLevel})

        x_filter = json.dumps(x_filter)

        result, result_code = self.getTable('View_Node', xFilter=x_filter, pageSize=pageSize, verbose=verbose)

        return result, result_code

    ######################################################################################
    def getAssetTypesID(self, assetTypes, pageSize=None, verbose=False):
        assetTypeIDs = []
        for assetType in assetTypes:
            at_x_filter = {"where": [{"field": "Name", "value": assetType}]}
            at_x_filter = json.dumps(at_x_filter)
            result, result_code = self.getTable('Comp_Type', xFilter=at_x_filter,
                                                pageSize=pageSize, verbose=verbose)

            if len(result['rows']) != 0:
                assetTypeIDs.append(result['rows'][0]['CT_ID'])
            else:
                assetTypeIDs.append('Asset type does not exist in the database')

        return assetTypeIDs

    ######################################################################################
    def getTableDBNames(self, tableNames, tableType, pageSize=None, verbose=False):
        tableDBNames = []

        for tableIndex, tableName in enumerate(tableNames):
            x_filter = {"operator": "and",
                        "where": [{"field": "Name", "value": tableName},
                                  {"field": "Def_Type.Name", "value": tableType}]
                        }
            x_filter = json.dumps(x_filter)

            result, result_code = self.getTable('Table_Def', xFilter=x_filter,
                                                pageSize=pageSize, verbose=verbose)

            for row in result['rows']:
                tableDBNames.append(row['Table_Name'])

        return tableDBNames

    ######################################################################################
    def getAssetByFullLocation(self, fullLocation, assetView=None, pageSize=None, verbose=False):
        """
            Get Asset details from View_Node table using full location.

            :param fullLocation: (``string``) - Full location is specified the same way it is defined in a typical NEXUS import sheet (single column)
            :param assetView: (``string`` - optional) - Asset view name (default value None)
            :param pageSize: (``int`` - optional) - Page size for response (default value None). When default is used the default NEXUS IC page size will be used (100).
            :param verbose: (``bool`` - optional) - Print internal messages if True (default value False).

            :return: (``tuple`` - dict or string, string) - response (json, raw or error string) and response status code.
        """

        x_filter = {"operator": "and",
                    "where": [{'field': 'View_Node.Full_Location', 'value': fullLocation}]
                    }

        # Apply asset view filter
        if assetView != None:
            x_filter['where'].append({"field": "Comp_View.Name", "value": assetView})

        x_filter = json.dumps(x_filter)

        result, result_code = self.getTable('View_Node', xFilter=x_filter,
                                            pageSize=pageSize, verbose=verbose)

        return result, result_code

    ######################################################################################
    def getTableDefInfo(self, tableName, tableType, pageSize=None, verbose=False):
        x_filter = {"operator": "and",
                    "where": [{"field": "Name", "value": tableName},
                              {"field": "Def_Type.Name", "value": tableType}]
                    }
        x_filter = json.dumps(x_filter)

        result, result_code = self.getTable('Table_Def', xFilter=x_filter,
                                            pageSize=pageSize, verbose=verbose)

        return result, result_code

    ######################################################################################
    def getLookupListItem(self, LI_ID, verbose=False):
        """
            Gets lookup item ifnormation given a lookup item ID

            :param LI_ID: (``int``) - Lookup item ID in the database
            :param verbose: (``bool`` - optional) - Print internal messages if True (default value False).

            :return: (``tuple`` - dict or string, string) - response (json, raw or error string) and response status code.
        """

        x_filter = {'operator': 'and',
                    'where': [{'field': 'Lookup_Item.LI_ID', 'value': LI_ID}]
                    }
        x_filter = json.dumps(x_filter)
        result, result_code = self.getTable('Lookup_Item', xFilter=x_filter, verbose=verbose)
        return result, result_code

    ######################################################################################
    def createNewEvents(self, events, sameAsset=False, sameWorkpack=False,
                        sameEventType=False, sameSurveySet=False,
                        pageSize=None, verbose=False):
        """
            Imports new events in the NEXUS IC database. This function creates new event and doesn't update existing events

            :param events: (``list``) - Events to be imported. Must follow the following format:

                .. code-block:: python

                        events = [{'component_id': xxxx,,
                                    'asset_full_location': 'xxxx / xxxx',
                                    'event_type': 'xxxx',
                                    'workpack_name': 'xxxx',
                                    'survey_set': 'xxxx',
                                    'start_clock': 'YYYY-MM-DDThh:mm:ss.ssssZ',
                                    'end_clock': 'YYYY-MM-DDThh:mm:ss.ssssZ',
                                    'comment': 'xxxxx',
                                    'fields': {'<Field Name 1>': xxx,
                                                '<Field Name 2>': xxx,
                                                '<Field Name 3>': xxx
                                                }
                                    }
                                ]

                where:

                - component_id: Optional if asset_full_location is provided - Using component_id provides better performance
                - asset_full_location: Optional if component_id is provided - Follows the same format as the standard NEXUS import sheet (single column)
                - survey_set: Optional (default value is Raw Survey Data)
                - start_clock and end_clock: YYYY for years, MM for months, DD for days, hh for hours, mm for minutes, ss for seconds and .ssss for milliseconds
                - fields: NEXUS field names are the keys (as shown in NEXUS IC)

            :param sameAsset: (``bool`` - optional) - Used when all events have the same asset.
            :param sameWorkpack: (``bool`` - optional) - Used when all events have the same workpack.
            :param sameEventType: (``bool`` - optional) - Used when all events have the same event type.
            :param sameSurveySet: (``bool`` - optional) - Used when all events have the same survey set.
            :param pageSize: (``int`` - optional) - Page size for response (default value None). When default is used the default NEXUS IC page size will be used (100).
            :param verbose: (``bool`` - optional) - Print internal messages if True (default value False).

            :return: (``tuple`` - list, list) - List of imported events and missed events.
        """

        imported_events = []
        missed_events = []
        prev_asset_full_location = None
        prev_workpack_name = None
        prev_event_type = None
        prev_survey_set = None

        for event_idx, event in enumerate(events):
            print('Importing new events...' + str(event_idx + 1) + ' of ' + str(len(events)))

            # Get Asset by Full Location
            # --------------------------
            if sameAsset and event_idx > 0:
                if component_id == None:
                    continue
            else:
                if 'component_id' in event.keys():
                    component_id = event['component_id']
                elif 'asset_full_location' in event.keys():
                    if event['asset_full_location'] != prev_asset_full_location:
                        prev_asset_full_location = event['asset_full_location']
                        result, result_code = self.getAssetByFullLocation(event['asset_full_location'], pageSize=pageSize,
                                                                          verbose=verbose)
                        if len(result['rows']) == 0:
                            component_id = None
                            missed_events.append({'Event': event,
                                                  'Reason': 'Asset was not found in database'})
                            continue
                        else:
                            component_id = result['rows'][0]['Component_ID']
                else:
                    component_id = None
                    missed_events.append({'Event': event,
                                          'Reason': 'asset_full_location key was not found in input dictionary'})
                    continue

            # Get Workpack ID
            # ---------------
            if sameWorkpack and event_idx > 0:
                if workpack_id == None:
                    continue
            else:
                if 'workpack_name' in event.keys():
                    if event['workpack_name'] != prev_workpack_name:
                        prev_workpack_name = event['workpack_name']
                        x_filter = {"operator": "and",
                                    "where": [{'field': 'Workpack.Name', 'value': event['workpack_name']}]
                                    }
                        x_filter = json.dumps(x_filter)
                        result, result_code = self.getTable('Workpack', xFilter=x_filter, pageSize=pageSize,
                                                            verbose=verbose)
                        if len(result['rows']) == 0:
                            workpack_id = None
                            missed_events.append({'Event': event,
                                                  'Reason': 'Workpack was not found in database'})
                            continue
                        else:
                            workpack_id = result['rows'][0]['Workpack_ID']
                else:
                    workpack_id = None
                    missed_events.append({'Event': event,
                                          'Reason': 'workpack_name key was not found in input dictionary'})
                    continue

            # Get event type table def information
            # ------------------------------------
            if sameEventType and event_idx > 0:
                if td_id == None:
                    continue
            else:
                if 'event_type' in event.keys():
                    if event['event_type'] != prev_event_type:
                        prev_event_type = event['event_type']
                        result, result_code = self.getTableDefInfo(event['event_type'], 'Event', pageSize=pageSize,
                                                                   verbose=verbose)

                        if len(result['rows']) == 0:
                            result, result_code = self.getTableDefInfo(event['event_type'], 'Continuous Event',
                                                                       pageSize=pageSize, verbose=verbose)

                            if len(result['rows']) == 0:
                                td_id = None
                                missed_events.append({'Event': event,
                                                      'Reason': 'Event type was not found in database'})
                                continue
                            else:
                                td_id = result['rows'][0]['TD_ID']
                                tableDBName = result['rows'][0]['Table_Name']
                        else:
                            td_id = result['rows'][0]['TD_ID']
                            tableDBName = result['rows'][0]['Table_Name']
                else:
                    td_id = None
                    missed_events.append({'Event': event,
                                          'Reason': 'event_type key was not found in input dictionary'})
                    continue

            # Get Survey Set ID
            # -----------------
            if sameSurveySet and event_idx > 0:
                if ss_id == None:
                    continue
            else:
                if 'survey_set' in event.keys():
                    survey_set = event['survey_set']
                else:
                    survey_set = 'Raw Survey Data'

                if prev_survey_set != survey_set:
                    prev_survey_set = survey_set
                    x_filter = {"operator": "and",
                                "where": [{'field': 'Survey_Set.Name', 'value': survey_set}]
                                }
                    x_filter = json.dumps(x_filter)
                    result, result_code = self.getTable('Survey_Set', xFilter=x_filter, pageSize=pageSize, verbose=verbose)

                    if (len(result['rows']) == 0) or (result_code != 200):
                        ss_id = None
                        missed_events.append({'Event': event,
                                              'Reason': 'can not find survey set in the database'})
                        continue

                    ss_id = result['rows'][0]['SS_ID']

            # Check if start_clock and end_clock is in input dictionary
            # ---------------------------------------------------------
            if not ('start_clock' in event.keys()):
                missed_events.append({'Event': event,
                                      'Reason': 'start_clock key was not found in input dictionary'})
                continue

            if not ('end_clock' in event.keys()):
                missed_events.append({'Event': event,
                                      'Reason': 'end_clock key was not found in input dictionary'})
                continue

            # Check event fields type and create event body
            # ---------------------------------------------
            event_body = {}
            for eventFieldName in event['fields'].keys():
                x_filter = {'operator': 'and',
                            'where': [{'field': 'Field_Def.Name', 'value': eventFieldName}]
                            }
                x_filter = json.dumps(x_filter)
                result, result_code = self.getTable('Field_Def', xFilter=x_filter, pageSize=pageSize,
                                                    verbose=verbose)

                if len(result['rows']) == 0:
                    if 'comment' in event.keys():
                        event['comment'] += '\nCheck this field: ' + eventFieldName + ': ' + str(
                            event['fields'][eventFieldName]) + ' '
                    else:
                        event['comment'] = eventFieldName + ': ' + str(event['fields'][eventFieldName]) + ' '
                else:
                    dbFieldName = result['rows'][0]['Field_Name']

                    # Check if field is lookup list type
                    if 'LL_ID' in result['rows'][0].keys():
                        x_filter = {'operator': 'and',
                                    'where': [{'field': 'Lookup_Item.LL_ID', 'value': result['rows'][0]['LL_ID']},
                                              {'field': 'Lookup_Item.Comments',
                                               'value': event['fields'][eventFieldName]}]
                                    }
                        x_filter = json.dumps(x_filter)
                        result, result_code = self.getTable('Lookup_Item', xFilter=x_filter, pageSize=pageSize,
                                                            verbose=verbose)

                        if len(result['rows']) == 0:
                            if 'comment' in event.keys():
                                event['comment'] += '\nCheck this field: ' + eventFieldName + ': ' + str(
                                    event['fields'][eventFieldName]) + ' '
                            else:
                                event['comment'] = eventFieldName + ': ' + str(
                                    event['fields'][eventFieldName]) + ' '
                        else:
                            event_body[dbFieldName] = result['rows'][0]['LI_ID']
                    else:
                        event_body[dbFieldName] = event['fields'][eventFieldName]

            # Create new header record
            # ------------------------
            header_body = {'TD_ID': td_id,
                           'Start_Clock': event['start_clock'],
                           'End_Clock': event['end_clock'],
                           'Workpack_ID': workpack_id,
                           'Component_ID': component_id,
                           'SS_ID': ss_id}

            result, result_code = self.createNewRecord('Header', header_body, verbose=verbose)
            event_body['Header_ID'] = result['rows'][0]['Header_ID']

            # Create new event record
            # -----------------------
            result, result_code = self.createNewRecord(tableDBName, event_body, verbose=verbose)

            # Create new commentary record
            # ----------------------------
            if 'comment' in event.keys():
                commentary_body = {'Header_ID': event_body['Header_ID'],
                                   'Notes': event['comment']}
                result, result_code = self.createNewRecord('Commentary', commentary_body, verbose=verbose)

            imported_events.append(event)

        print('Imported ' + str(len(imported_events)) + ' new events...')
        print('Missed ' + str(len(missed_events)) + ' events from input...')
        return imported_events, missed_events

######################################################################################
################################### Start Script #####################################
if __name__ == '__main__':
    baseURI = ''
    apiKey = ''

    startTime = datetime.datetime.now()

    ## Program start here
    print(NEXUSIC_REST.__doc__)
    ## End of program

    endTime = datetime.datetime.now()
    elapsedTime = endTime - startTime

    print('NEXUS IC REST API actions completed.....runtime: %s' % (str(elapsedTime)))





import requests
import json
from requests.auth import HTTPBasicAuth
#from requests_kerberos import HTTPKerberosAuth
from requests_ntlm import HttpNtlmAuth

############################# OSI PI Web API Class ############################
class OSIPI_WebAPI():
    """
    OSI PI Web API class allows the user to communicate with OSI PI web API using python

    Prerequisites:
    -------------
        - Python > 3.7
        - OSI PI access

    OSI PI Documentation:
    --------------------
        The OSI PI Web API documentation can be found in the bellow link:
        https://techsupport.osisoft.com/Documentation/PI-Web-API/help.html
    """

    def __init__(self, piwebapi_url, username, password, authentication_type='basic',
                 max_attempts=1, timeout=None,
                 verbose=False, verify=True):
        """

            :param piwebapi_url: (``string``) - OSI PI base URL.
            :param username: (``string``) - OSI PI username.
            :param password: (``string``) - OSI PI password.
            :param authentication_type: (``string`` - optional) - This can be either ntlm or basic (default value basic).
            :param max_attempts: (``int`` - optional) - Maximum number of attempts if disconnected (default value 1).
            :param timeout: (``int`` - optional) - Timeout threshold in seconds (default value None).
            :param verbose: (``bool`` - optional) - Print internal messages if True (default value False).
            :param verify: (``bool`` - optional) - By pass SSL verification if True (default value True).

            :returns: None
        """
        assert authentication_type.lower() in ['ntlm', 'basic'], 'Incorrect security method'
        self.piwebapi_url = piwebapi_url
        self.authentication_type = authentication_type
        self.username = username
        self.password = password
        self.max_attempts = max_attempts   # TODO: Functionality to be added
        self.timeout = timeout  # TODO: Functionality to be added
        self.verbose = verbose
        self.verify = verify

        self.authentication_type = self.authenticate(self.authentication_type, self.username, self.password)
        self.sys_links = self.get_system_links()

    ########################################################################################################
    def authenticate(self, authentication_type, username, password, verbose=False):
        """
            Authenticate and create API call security method

            :param authentication_type: (``string``) - Authentication method to use:  basic or ntlm.
            :param username: (``string``) - The user's credentials name.
            :param password: (``string``) - The user's credentials password.
            :param verbose: (``bool`` - optional) - Print internal messages if True (default value False).

            :returns: (``HttpNtlmAuth`` or ``HTTpBasicAuth``)
        """
        assert authentication_type.lower() in ['basic', 'ntlm'], authentication_type + ' authentication type not supported'

        if self.verbose or verbose:
            print('Authenticating with OSI PI...')

        if authentication_type.lower() == 'basic':
            security_auth = HTTPBasicAuth(username, password)
        elif authentication_type.lower() == 'ntlm':
            security_auth = HttpNtlmAuth(username, password)
        #else:
        #    security_auth = HTTPKerberosAuth(mutual_authentication='REQUIRED',
        #                                     sanitize_mutual_error_response=False)

        return security_auth

    ######################################################################################
    def validate_and_return_response(self, response, message='', raw=False):
        """
            Validate response by comparing it against the acceptable status codes. Returns response and status code.

            :param response: (``requests.response``) - The response from the query.
            :param message: (``string`` - optional) - Desired custom error message.
            :param raw: (``bool`` - optional) - Defines whether to return response raw format or json (default value False).

            :return: (``tuple`` - dict or string, string) - response (json, raw or error string) and response status code.
        """

        if response.status_code in [200]:
            if raw:
                return response.raw, response.status_code
            else:
                return response.json(), response.status_code
        else:
            return str(message) + str(response.status_code) + ': ' + str(response.text), \
                   response.status_code

    ########################################################################################################
    def webapi_get_request(self, url, pageSize=None, message='', raw=False):
        """
            Execute get request and validate response.

            :param url: (``string``) - URL to be executed in the get request.
            :param pageSize: (``int`` - optional) - Page size for response (default value None). When default is used the default OSI PI page size will be used (1000).
            :param message: (``string`` - optional) - Desired custom error message.
            :param raw: (``bool`` - optional) - Defines whether to return response raw format or json (default value False).

            :return: (``tuple`` - dict or string, string) - response (json, raw or error string) and response status code.
        """

        if pageSize != None:
            url = url + '/?maxCount=' + str(pageSize)

        response = requests.get(url, auth=self.authentication_type, verify=self.verify)

        resp, resp_status = self.validate_and_return_response(response, message=message, raw=raw)

        return resp, resp_status

    ########################################################################################################
    def get_system_links(self, pageSize=None, verbose=False):
        """
            Get OSI PI DataServers and AssetServers available databases

            :param pageSize: (``int`` - optional) - Page size for response (default value None). When default is used the default OSI PI page size will be used (1000).
            :param verbose: (``bool`` - optional) - Print internal messages if True (default value False).

            :return: (``dict``) - System links dictionary

                .. code-block:: python

                    {'<serverType>': [{'Name': 'xxx', 'Link': 'xxx', 'WebID': 'xxx'}]}

                **<serverType>** is either DataServers or AssetServers
        """

        if self.verbose or verbose:
            print('Getting OSI PI system links...')

        result = {}
        links, links_status = self.webapi_get_request(self.piwebapi_url, pageSize=pageSize)
        for link in links['Links']:
            if link in ['DataServers', 'AssetServers']:
                response, resp_status = self.webapi_get_request(links['Links'][link], pageSize=pageSize)
                itemsList = []
                for item in response['Items']:
                    itemsList.append({'Name': item['Name'],
                                    'Link': links['Links'][link],
                                    'WebID': item['WebId']})
                result[link] = itemsList
        return result

    ########################################################################################################
    def getPointsList(self, dataServerName=None, pageSize=None, verbose=False):
        """
            Get full point list (sensors) in a specified data server. If data server is None, then all points from all data servers will be returned

            :param dataServerName: (``string`` - optional) - Specific data server to retrieve data from (default value None).
            :param pageSize: (``int`` - optional) - Page size for response (default value None). When default is used the default OSI PI page size will be used (1000).
            :param verbose: (``bool`` - optional) - Print internal messages if True (default value False).

            :return: (``tuple`` - dict, None) - List of all points (sensors). The second argument of the tuple is always None

                .. code-block:: python

                    ([{'ServerName': 'xxx', 'Points': 'xxx', 'ResponseStatus': 'xxx'}], None)
        """

        if self.verbose or verbose:
            print('Getting OSI PI points list...')

        result = []
        for dataServer in self.sys_links['DataServers']:
            if dataServerName != None:
                if dataServerName != dataServer['Name']:
                    continue

            points_url = dataServer['Link'] + r'/' + \
                         dataServer['WebID'] + r'/points'

            points_resp, points_resp_status = self.webapi_get_request(points_url, pageSize=pageSize)

            if points_resp_status == 200:
                points = points_resp['Items']
            else:
                points = points_resp

            result.append({'ServerName': dataServer['Name'],
                           'Points': points,
                           'ResponseStatus': points_resp_status})

        return result, None

    ########################################################################################################
    def getStreamData(self, webID, searchType='end', parameters=None, pageSize=None, verbose=False):
        """
            Get stream data for a specified point (sensor) using point's webID. Search filter can be applied to narrow down the result

            :param webID: (``string``) - Point (sensor) web ID.
            :param searchType: (``string`` - optional) - Search type to be used, can be one of the following values (default value ``end``).

                .. code-block:: python

                        ['end', 'interpolated', 'recorded', 'plot', 'summary', 'value']

            :param parameters: (``string`` - optional) - Additional parameters to filter the query (default value None).
                See https://techsupport.osisoft.com/Documentation/PI-Web-API/help/controllers/stream.html for additional information for each search type.
            :param pageSize: (``int`` - optional) - Page size for response (default value None). When default is used the default OSI PI page size will be used (1000).
            :param verbose: (``bool`` - optional) - Print internal messages if True (default value False).

            :return: (``tuple`` - dict or string, string) - response (json or error string) and response status code.
        """

        assert searchType.lower() in ['end', 'interpolated', 'recorded',
                                      'plot', 'summary', 'value'], searchType + ' search type not supported'

        if self.verbose or verbose:
            print('Getting OSI PI stream data...')

        if parameters == None:
            url = self.piwebapi_url + r'/streams/' + str(webID) + r'/' + searchType.lower()
        else:
            url = self.piwebapi_url + r'/streams/' + str(webID) + r'/' + searchType.lower() + '?' + parameters

        resp, resp_status = self.webapi_get_request(url, pageSize=pageSize)
        return resp, resp_status

    ########################################################################################################
    def getStreamDataSummary(self, webID, startTime=None, endTime=None, summaryType=None,
                             pageSize=None, verbose=False):
        """
            Get stream data summary for a specified point (sensor) using point's webID. Search filter can be applied to narrow down the result.

            For more details see https://techsupport.osisoft.com/Documentation/PI-Web-API/help/controllers/stream/actions/getsummary.html

            :param webID: (``string``) - Point (sensor) web ID.
            :param startTime: (``string`` - optional) -Start time filter (default vale None)
            :param endTime: (``string`` - optional) -End time filter (default vale None)
            :param summaryType: (``string`` - optional) - Summary type requested, can be one of the following values (default value None).

                .. code-block:: python

                        ['Total', 'Average', 'Minimum', 'Maximum', 'Range',
                        'StdDev', 'PopulationStdDev', 'Count', 'PercentGood',
                        'TotalWithUOM', 'All', 'AllForNonNumeric']

            :param pageSize: (``int`` - optional) - Page size for response (default value None). When default is used the default OSI PI page size will be used (1000).
            :param verbose: (``bool`` - optional) - Print internal messages if True (default value False).

            :return: (``tuple`` - dict or string, string) - response (json or error string) and response status code.
        """

        assert summaryType in ['Total', 'Average', 'Minimum', 'Maximum', 'Range',
                               'StdDev', 'PopulationStdDev', 'Count', 'PercentGood',
                               'TotalWithUOM', 'All', 'AllForNonNumeric'], summaryType + ' summary type not supported'
        parameters = None
        if startTime != None:
            parameters = 'startTime=' + str(startTime)

        if endTime != None:
            if parameters == None:
                parameters = 'endTime=' + str(endTime)
            else:
                parameters += '&endTime=' + str(endTime)

        if summaryType != None:
            if parameters == None:
                parameters = 'summaryType=' + str(summaryType)
            else:
                parameters += '&summaryType=' + str(summaryType)

        resp, resp_status = self.getStreamData(webID, searchType='summary', parameters=parameters,
                                               pageSize=pageSize, verbose=verbose)
        return resp, resp_status

    ########################################################################################################
    def getPointAttributes(self, webID, pageSize=None, verbose=False):
        """
            Get point (sensor) system attributes

            For more details see https://techsupport.osisoft.com/Documentation/PI-Web-API/help/controllers/point/actions/getattributes.html

            :param webID: (``string``) - Point (sensor) web ID.
            :param pageSize: (``int`` - optional) - Page size for response (default value None). When default is used the default OSI PI page size will be used (1000).
            :param verbose: (``bool`` - optional) - Print internal messages if True (default value False).

            :return: (``tuple`` - dict or string, string) - response (json or error string) and response status code.
        """

        if self.verbose or verbose:
            print('Getting OSI PI point attributes...')

        url = self.piwebapi_url + r'/points/' + str(webID) + r'/attributes'
        resp, resp_status = self.webapi_get_request(url, pageSize=pageSize)
        return resp, resp_status

############################################################################################################
############################################################################################################
if __name__ == '__main__':
    pass


import requests
import base64
import datetime
import traceback
import json

############################# Maximo REST API Class ############################
class MAXIMO_REST():
    """
        Maximo API class allows the user to communicate with Maximo REST API using python

        Prerequisites:
        -------------
            - Python > 3.7
            - Maximo access
    """


    _error_msgs = ['An existing connection was forcibly closed by the remote host',
                   'Remote end closed connection without response']

    def __init__(self, base_url, username=None, password=None, api_key=None,
                 authentication_type='LDAP BASIC',
                 max_attempts=1, timeout=None,
                 verbose=False, verify=True):
        """

            :param maximo_url: (``string``) - Maximo URL
            :param username: (``string`` - optional) - Default value None
            :param password: (``string`` - optional) - Default value None
            :param api_key: (``string`` - optional) - Default value None
            :param authentication_type: (``string`` - optional) - The only value supported LDAP BASIC (default value LDAP BASIC)
            :param max_attempts: (``int`` - optional) - Maximum number of attempts if disconnected (default value 1)
            :param timeout: (``int`` - optional) - Timeout threshold in seconds (default value None)
            :param verbose: (``bool`` - optional) - Print internal messages if True (default value False)
            :param verify: (``bool`` - optional) - By pass SSL verification if True (default value True)

            :returns: None
        """

        assert authentication_type in ['LDAP BASIC'], 'Incorrect authentication type'

        if username == None:
            raise Exception('Username is empty, please provide a valid username')

        if password == None:
            raise Exception('Password is empty, please provide a valid password')

        self.base_url = base_url
        self.authentication_type = authentication_type
        self.username = username
        self.password = password
        self.api_key = api_key
        self.max_attempts = max_attempts   # TODO: Functionality to be added
        self.timeout = timeout   # TODO: To be added to the REST calls
        self.verbose = verbose
        self.verify = verify
        self.api_key = self.generate_base64(self.username + ':' + self.password)
        self.session = None
        self.authenticate()

    ######################################################################################
    def generate_base64(self, value):
        """
            Generate base64 string

            :param value: (``string``) - String value to be converted to base64.

            :returns: (``String``) - base64 string.
        """

        return str(base64.b64encode(bytes(value, 'utf-8')), "utf-8")

    ######################################################################################
    def validate_and_return_response(self, response, message, raw=False):
        """
            Validate response by comparing it against the acceptable status codes. Returns response and status code.

            :param response: (``requests.response``) - The response from the query.
            :param message: (``string`` - optional) - Desired custom error message.
            :param raw: (``bool`` - optional) - Defines whether to return response raw format or json (default value False).

            :return: (``tuple`` - dict or string, string) - response (json, raw or error string) and response status code.
        """

        if response.status_code in [200, 201, 204]:
            if raw:
                return response.raw, response.status_code
            else:
                return response.json(), response.status_code
        else:
            return str(message) + str(response.status_code) + ': ' + str(response.text), \
                   response.status_code

    ######################################################################################
    def authenticate(self, verbose=False):
        """
            Authenticate with Maximo using the defined authentication type used in the class constructor

            :param verbose: (``bool`` - optional) - Print internal messages if True (default value False).

            :returns: None
        """

        if self.verbose or verbose:
            print('Authenticating with Maximo...')

        if self.authentication_type == 'LDAP BASIC':
            url = self.base_url + '/maximo/j_security_check?j_username=' + \
                  self.username + '&j_password=' + self.password

            self.session = requests.Session()
            self.session.get(url)
        else:
            raise Exception(self.authentication_type + ' authentication type not supported')

    ######################################################################################
    def mxapi_get_request(self, url):
        """
            Execute get request.

            :param url: (``string``) - URL to be executed in the get request.

            :return: (``requests.response``)
        """

        if self.authentication_type == 'LDAP BASIC':
            return self.session.get(url,
                                    headers={'Authorization': 'Basic' + self.api_key},
                                    verify=self.verify)
        else:
            raise Exception(self.authentication_type + ' authentication type not supported')

    ######################################################################################
    def get_table(self, tableName, pageSize=None, current_attempt=1,
                 verbose=False, select=None, where=None, orderBy=None,
                 pageNum=1, stablePaging=False):
        """
            Get data from a specific table in Maximo.

            :param tableName: (``string``) - Maximo table name to extract data
            :param pageSize: (``int`` - optional) - Page size for response (default value None). When default is used the default Maximo page size will be used (1000).
            :param current_attempt: (``int`` - **don't use**) - This arugment is intended for internal method use only.
            :param verbose: (``bool`` - optional) - Print internal messages if True (default value False).
            :param select: (``string`` - optional) - Maximo table fields to be included in the select query (default value None).
            :param where: (``string`` - optional) - Where condition to filter data (default value None).
            :param orderBy: (``string`` - optional) - Field(s) to order data (default value None).
            :param pageNum: (``int`` - optional) - Page number in the response (default value 1). Recommended to leave default.
            :param stablePaging: (``bool`` - optional) - Use Maximo stable paging option (default value False). Recommended to leave default.

            :return: (``tuple`` - dict or string, string) - response (json or error string) and response status code.
        """

        if self.verbose or verbose:
            print('Getting ' + str(tableName) + '...')

        # Adding /oslc/ part to the baseURI
        url = self.base_url + '/maximo/oslc/os/' + tableName

        # Check page size
        if pageSize != None:
            if '?' in url:
                url += '&oslc.pageSize=' + str(pageSize)
            else:
                url += '?oslc.pageSize=' + str(pageSize)

        # Add select query
        if select != None:
            if '?' in url:
                url += '&oslc.select=' + str(select)
            else:
                url += '?oslc.select=' + str(select)

        # Add where query
        if where != None:
            if '?' in url:
                url += '&oslc.where=' + str(where)
            else:
                url += '?oslc.where=' + str(where)

        # Add order by query
        if orderBy != None:
            if '?' in url:
                url += '&oslc.orderBy=+' + str(orderBy)
            else:
                url += '?oslc.orderBy=+' + str(orderBy)

        # Check stable paging
        if stablePaging:
            if '?' in url:
                url += '&stablepaging=true'
            else:
                url += '?stablepaging=true'

        # Check pageNum
        if pageNum > 1:
            if '?' in url:
                url += '&pageno=' + str(pageNum)
            else:
                url += '?pageno=' + str(pageNum)

        # Execute request
        has_more_records = True
        result = {'members': []}
        try:
            while has_more_records:
                response = self.mxapi_get_request(url)
                resp_result, resp_code = self.validate_and_return_response(response,
                                                                           'Get ' + tableName + ' table error')
                result['members'].extend(resp_result['rdfs:member'])

                if 'oslc:nextPage' in resp_result['oslc:responseInfo']:
                    prev_pageNum = pageNum
                    pageNum += 1
                    if '&pageno=' in url:
                        url = url.replace('&pageno=' + str(prev_pageNum), '&pageno=' + str(pageNum))
                    else:
                        url += '&pageno=' + str(pageNum)
                else:
                    has_more_records = False

            return result, resp_code
        except Exception as e:
            for error_msg in MAXIMO_REST._error_msgs:
                if error_msg in str(e):
                    current_attempt += 1
                    if current_attempt <= self.max_attempts:
                        self.authenticate()
                        return self.get_table(tableName, pageSize=pageSize,
                                              current_attempt=current_attempt,
                                              verbose=verbose, select=select,
                                              where=where, orderBy=orderBy,
                                              pageNum=pageNum,
                                              stablePaging=stablePaging)

            errorMsg = traceback.format_exc()
            raise Exception('Error getting data from table: ' + str(tableName) + '\n' + errorMsg)

######################################################################################
################################### Start Script #####################################
if __name__ == '__main__':
    maximoURL = ''
    username = ''
    password = ''

    mx_obj = MAXIMO_REST(maximoURL, username, password)
    print(mx_obj.getTable('mxasset', select='*'))